"use strict";
var __extends = (this && this.__extends) || (function () {
    var extendStatics = Object.setPrototypeOf ||
        ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
        function (d, b) { for (var p in b) if (b.hasOwnProperty(p)) d[p] = b[p]; };
    return function (d, b) {
        extendStatics(d, b);
        function __() { this.constructor = d; }
        d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
    };
})();
Object.defineProperty(exports, "__esModule", { value: true });
var immutable_1 = require("immutable");
var TargetLanguage_1 = require("../TargetLanguage");
var RendererOptions_1 = require("../RendererOptions");
var Type_1 = require("../Type");
var ConvenienceRenderer_1 = require("../ConvenienceRenderer");
var Naming_1 = require("../Naming");
var Strings_1 = require("../Strings");
var Support_1 = require("../Support");
var Source_1 = require("../Source");
var Annotation_1 = require("../Annotation");
var ElmTargetLanguage = /** @class */ (function (_super) {
    __extends(ElmTargetLanguage, _super);
    function ElmTargetLanguage() {
        var _this = this;
        var listOption = new RendererOptions_1.EnumOption("array-type", "Use Array or List", [["array", false], ["list", true]]);
        // FIXME: Do this via a configurable named eventually.
        var moduleOption = new RendererOptions_1.StringOption("module", "Generated module name", "NAME", "QuickType");
        var options = [moduleOption, listOption];
        _this = _super.call(this, "Elm", ["elm"], "elm", options.map(function (o) { return o.definition; })) || this;
        _this._listOption = listOption;
        _this._moduleOption = moduleOption;
        return _this;
    }
    ElmTargetLanguage.prototype.renderGraph = function (graph, optionValues) {
        var renderer = new ElmRenderer(graph, this._listOption.getValue(optionValues), this._moduleOption.getValue(optionValues));
        return renderer.render();
    };
    return ElmTargetLanguage;
}(TargetLanguage_1.TargetLanguage));
exports.default = ElmTargetLanguage;
var forbiddenNames = [
    "if",
    "then",
    "else",
    "case",
    "of",
    "let",
    "in",
    "type",
    "module",
    "where",
    "import",
    "exposing",
    "as",
    "port",
    "int",
    "float",
    "bool",
    "string",
    "Jenc",
    "Jdec",
    "Jpipe",
    "always",
    "identity",
    "Array",
    "List",
    "Dict",
    "Maybe",
    "map",
    "toList",
    "makeArrayEncoder",
    "makeDictEncoder",
    "makeNullableEncoder"
];
var legalizeName = Strings_1.legalizeCharacters(function (cp) { return Strings_1.isAscii(cp) && Strings_1.isLetterOrUnderscoreOrDigit(cp); });
function elmNameStyle(original, upper) {
    var legalized = legalizeName(original);
    var pascaled = Strings_1.pascalCase(legalized);
    var result = Strings_1.startWithLetter(Strings_1.isLetterOrUnderscore, upper, pascaled);
    return result;
}
var upperNamingFunction = Naming_1.funPrefixNamer(function (n) { return elmNameStyle(n, true); });
var lowerNamingFunction = Naming_1.funPrefixNamer(function (n) { return elmNameStyle(n, false); });
function singleWord(source) {
    return { source: source, needsParens: false };
}
function multiWord(a, b) {
    return { source: [a, " ", b], needsParens: true };
}
function parenIfNeeded(_a) {
    var source = _a.source, needsParens = _a.needsParens;
    if (needsParens) {
        return ["(", source, ")"];
    }
    return source;
}
function requiredOrOptional(t) {
    function optional(fallback) {
        return { reqOrOpt: "Jpipe.optional", fallback: fallback };
    }
    if (t.kind === "null") {
        return optional(" ()");
    }
    if (t instanceof Type_1.UnionType && Type_1.nullableFromUnion(t)) {
        return optional(" Nothing");
    }
    return { reqOrOpt: "Jpipe.required", fallback: "" };
}
var ElmRenderer = /** @class */ (function (_super) {
    __extends(ElmRenderer, _super);
    function ElmRenderer(graph, _useList, _moduleName) {
        var _this = _super.call(this, graph) || this;
        _this._useList = _useList;
        _this._moduleName = _moduleName;
        _this._topLevelDependents = immutable_1.Map();
        _this._namedTypeDependents = immutable_1.Map();
        _this.elmType = function (t, withIssues) {
            if (withIssues === void 0) { withIssues = false; }
            return Type_1.matchType(t, function (anyType) { return singleWord(Source_1.maybeAnnotated(withIssues, Annotation_1.anyTypeIssueAnnotation, "Jdec.Value")); }, function (nullType) { return singleWord(Source_1.maybeAnnotated(withIssues, Annotation_1.nullTypeIssueAnnotation, "()")); }, function (boolType) { return singleWord("Bool"); }, function (integerType) { return singleWord("Int"); }, function (doubleType) { return singleWord("Float"); }, function (stringType) { return singleWord("String"); }, function (arrayType) { return multiWord(_this.arrayType, parenIfNeeded(_this.elmType(arrayType.items, withIssues))); }, function (classType) { return singleWord(_this.nameForNamedType(classType)); }, function (mapType) { return multiWord("Dict String", parenIfNeeded(_this.elmType(mapType.values, withIssues))); }, function (enumType) { return singleWord(_this.nameForNamedType(enumType)); }, function (unionType) {
                var nullable = Type_1.nullableFromUnion(unionType);
                if (nullable)
                    return multiWord("Maybe", parenIfNeeded(_this.elmType(nullable, withIssues)));
                return singleWord(_this.nameForNamedType(unionType));
            });
        };
        _this.decoderNameForNamedType = function (t) {
            var name = _this.nameForNamedType(t);
            return Support_1.defined(_this._namedTypeDependents.get(name)).decoder;
        };
        _this.decoderNameForType = function (t) {
            return Type_1.matchType(t, function (anyType) { return singleWord("Jdec.value"); }, function (nullType) { return multiWord("Jdec.null", "()"); }, function (boolType) { return singleWord("Jdec.bool"); }, function (integerType) { return singleWord("Jdec.int"); }, function (doubleType) { return singleWord("Jdec.float"); }, function (stringType) { return singleWord("Jdec.string"); }, function (arrayType) {
                return multiWord(["Jdec.", Strings_1.decapitalize(_this.arrayType)], parenIfNeeded(_this.decoderNameForType(arrayType.items)));
            }, function (classType) { return singleWord(_this.decoderNameForNamedType(classType)); }, function (mapType) { return multiWord("Jdec.dict", parenIfNeeded(_this.decoderNameForType(mapType.values))); }, function (enumType) { return singleWord(_this.decoderNameForNamedType(enumType)); }, function (unionType) {
                var nullable = Type_1.nullableFromUnion(unionType);
                if (nullable)
                    return multiWord("Jdec.nullable", parenIfNeeded(_this.decoderNameForType(nullable)));
                return singleWord(_this.decoderNameForNamedType(unionType));
            });
        };
        _this.encoderNameForNamedType = function (t) {
            var name = _this.nameForNamedType(t);
            return Support_1.defined(_this._namedTypeDependents.get(name)).encoder;
        };
        _this.encoderNameForType = function (t) {
            return Type_1.matchType(t, function (anyType) { return singleWord("identity"); }, function (nullType) { return multiWord("always", "Jenc.null"); }, function (boolType) { return singleWord("Jenc.bool"); }, function (integerType) { return singleWord("Jenc.int"); }, function (doubleType) { return singleWord("Jenc.float"); }, function (stringType) { return singleWord("Jenc.string"); }, function (arrayType) {
                return multiWord(["make", _this.arrayType, "Encoder"], parenIfNeeded(_this.encoderNameForType(arrayType.items)));
            }, function (classType) { return singleWord(_this.encoderNameForNamedType(classType)); }, function (mapType) { return multiWord("makeDictEncoder", parenIfNeeded(_this.encoderNameForType(mapType.values))); }, function (enumType) { return singleWord(_this.encoderNameForNamedType(enumType)); }, function (unionType) {
                var nullable = Type_1.nullableFromUnion(unionType);
                if (nullable)
                    return multiWord("makeNullableEncoder", parenIfNeeded(_this.encoderNameForType(nullable)));
                return singleWord(_this.encoderNameForNamedType(unionType));
            });
        };
        _this.unionConstructorName = function (unionName, t) {
            return [elmNameStyle(_this.unionFieldName(t), true), "In", unionName];
        };
        _this.emitTopLevelDefinition = function (t, topLevelName) {
            _this.emitLine("type alias ", topLevelName, " = ", _this.elmType(t).source);
        };
        _this.emitClassDefinition = function (c, className) {
            _this.emitLine("type alias ", className, " =");
            _this.indent(function () {
                var onFirst = true;
                _this.forEachProperty(c, "none", function (name, _, t) {
                    _this.emitLine(onFirst ? "{" : ",", " ", name, " : ", _this.elmType(t).source);
                    onFirst = false;
                });
                if (onFirst) {
                    _this.emitLine("{");
                }
                _this.emitLine("}");
            });
        };
        _this.emitEnumDefinition = function (e, enumName) {
            _this.emitLine("type ", enumName);
            _this.indent(function () {
                var onFirst = true;
                _this.forEachCase(e, "none", function (name) {
                    var equalsOrPipe = onFirst ? "=" : "|";
                    _this.emitLine(equalsOrPipe, " ", name);
                    onFirst = false;
                });
            });
        };
        _this.emitUnionDefinition = function (u, unionName) {
            var members = u.members.sortBy(_this.unionFieldName);
            _this.emitLine("type ", unionName);
            _this.indent(function () {
                var onFirst = true;
                members.forEach(function (t) {
                    var equalsOrPipe = onFirst ? "=" : "|";
                    var constructor = _this.unionConstructorName(unionName, t);
                    if (t.kind === "null") {
                        _this.emitLine(equalsOrPipe, " ", constructor);
                    }
                    else {
                        _this.emitLine(equalsOrPipe, " ", constructor, " ", parenIfNeeded(_this.elmType(t)));
                    }
                    onFirst = false;
                });
            });
        };
        _this.emitTopLevelFunctions = function (t, topLevelName) {
            var _a = Support_1.defined(_this._topLevelDependents.get(topLevelName)), encoder = _a.encoder, decoder = _a.decoder;
            if (!_this.namedTypeToNameForTopLevel(t)) {
                _this.emitLine(Support_1.defined(decoder), " : Jdec.Decoder ", topLevelName);
                _this.emitLine(Support_1.defined(decoder), " = ", _this.decoderNameForType(t).source);
                _this.emitNewline();
            }
            _this.emitLine(encoder, " : ", topLevelName, " -> String");
            _this.emitLine(encoder, " r = Jenc.encode 0 (", _this.encoderNameForType(t).source, " r)");
        };
        _this.emitClassFunctions = function (c, className) {
            var decoderName = _this.decoderNameForNamedType(c);
            _this.emitLine(decoderName, " : Jdec.Decoder ", className);
            _this.emitLine(decoderName, " =");
            _this.indent(function () {
                _this.emitLine("Jpipe.decode ", className);
                _this.indent(function () {
                    _this.forEachProperty(c, "none", function (name, jsonName, t) {
                        var propDecoder = parenIfNeeded(_this.decoderNameForType(t));
                        var _a = requiredOrOptional(t), reqOrOpt = _a.reqOrOpt, fallback = _a.fallback;
                        _this.emitLine("|> ", reqOrOpt, ' "', Strings_1.stringEscape(jsonName), '" ', propDecoder, fallback);
                    });
                });
            });
            _this.emitNewline();
            var encoderName = _this.encoderNameForNamedType(c);
            _this.emitLine(encoderName, " : ", className, " -> Jenc.Value");
            _this.emitLine(encoderName, " x =");
            _this.indent(function () {
                _this.emitLine("Jenc.object");
                _this.indent(function () {
                    var onFirst = true;
                    _this.forEachProperty(c, "none", function (name, jsonName, t) {
                        var bracketOrComma = onFirst ? "[" : ",";
                        var propEncoder = _this.encoderNameForType(t).source;
                        _this.emitLine(bracketOrComma, ' ("', Strings_1.stringEscape(jsonName), '", ', propEncoder, " x.", name, ")");
                        onFirst = false;
                    });
                    if (onFirst) {
                        _this.emitLine("[");
                    }
                    _this.emitLine("]");
                });
            });
        };
        _this.emitEnumFunctions = function (e, enumName) {
            var decoderName = _this.decoderNameForNamedType(e);
            _this.emitLine(decoderName, " : Jdec.Decoder ", enumName);
            _this.emitLine(decoderName, " =");
            _this.indent(function () {
                _this.emitLine("Jdec.string");
                _this.indent(function () {
                    _this.emitLine("|> Jdec.andThen (\\str ->");
                    _this.indent(function () {
                        _this.emitLine("case str of");
                        _this.indent(function () {
                            _this.forEachCase(e, "none", function (name, jsonName) {
                                _this.emitLine('"', Strings_1.stringEscape(jsonName), '" -> Jdec.succeed ', name);
                            });
                            _this.emitLine('somethingElse -> Jdec.fail <| "Invalid ', enumName, ': " ++ somethingElse');
                        });
                    });
                    _this.emitLine(")");
                });
            });
            _this.emitNewline();
            var encoderName = _this.encoderNameForNamedType(e);
            _this.emitLine(encoderName, " : ", enumName, " -> Jenc.Value");
            _this.emitLine(encoderName, " x = case x of");
            _this.indent(function () {
                _this.forEachCase(e, "none", function (name, jsonName) {
                    _this.emitLine(name, ' -> Jenc.string "', Strings_1.stringEscape(jsonName), '"');
                });
            });
        };
        _this.emitUnionFunctions = function (u, unionName) {
            // We need arrays first, then strings.
            function sortOrder(t) {
                if (t.kind === "array") {
                    return "  array";
                }
                else if (t instanceof Type_1.PrimitiveType) {
                    return " " + t.kind;
                }
                return t.kind;
            }
            var members = u.members.sortBy(sortOrder);
            var decoderName = _this.decoderNameForNamedType(u);
            _this.emitLine(decoderName, " : Jdec.Decoder ", unionName);
            _this.emitLine(decoderName, " =");
            _this.indent(function () {
                _this.emitLine("Jdec.oneOf");
                _this.indent(function () {
                    var onFirst = true;
                    members.forEach(function (t) {
                        var bracketOrComma = onFirst ? "[" : ",";
                        var constructor = _this.unionConstructorName(unionName, t);
                        if (t.kind === "null") {
                            _this.emitLine(bracketOrComma, " Jdec.null ", constructor);
                        }
                        else {
                            var decoder = parenIfNeeded(_this.decoderNameForType(t));
                            _this.emitLine(bracketOrComma, " Jdec.map ", constructor, " ", decoder);
                        }
                        onFirst = false;
                    });
                    _this.emitLine("]");
                });
            });
            _this.emitNewline();
            var encoderName = _this.encoderNameForNamedType(u);
            _this.emitLine(encoderName, " : ", unionName, " -> Jenc.Value");
            _this.emitLine(encoderName, " x = case x of");
            _this.indent(function () {
                members.forEach(function (t) {
                    var constructor = _this.unionConstructorName(unionName, t);
                    if (t.kind === "null") {
                        _this.emitLine(constructor, " -> Jenc.null");
                    }
                    else {
                        var encoder = _this.encoderNameForType(t).source;
                        _this.emitLine(constructor, " y -> ", encoder, " y");
                    }
                });
            });
        };
        return _this;
    }
    Object.defineProperty(ElmRenderer.prototype, "forbiddenNamesForGlobalNamespace", {
        get: function () {
            return forbiddenNames;
        },
        enumerable: true,
        configurable: true
    });
    ElmRenderer.prototype.topLevelNameStyle = function (rawName) {
        return elmNameStyle(rawName, true);
    };
    ElmRenderer.prototype.topLevelDependencyNames = function (t, topLevelName) {
        var encoder = new Naming_1.DependencyName(lowerNamingFunction, immutable_1.List([topLevelName]), function (names) { return Strings_1.decapitalize(Support_1.defined(names.first())) + "ToString"; });
        var decoder = undefined;
        if (!this.namedTypeToNameForTopLevel(t)) {
            decoder = new Naming_1.DependencyName(lowerNamingFunction, immutable_1.List([topLevelName]), function (names) { return Support_1.defined(names.first()); });
        }
        this._topLevelDependents = this._topLevelDependents.set(topLevelName, { encoder: encoder, decoder: decoder });
        if (decoder !== undefined) {
            return [encoder, decoder];
        }
        return [encoder];
    };
    Object.defineProperty(ElmRenderer.prototype, "namedTypeNamer", {
        get: function () {
            return upperNamingFunction;
        },
        enumerable: true,
        configurable: true
    });
    ElmRenderer.prototype.namedTypeDependencyNames = function (t, typeName) {
        var encoder = new Naming_1.DependencyName(lowerNamingFunction, immutable_1.List([typeName]), function (names) { return "encode" + Support_1.defined(names.first()); });
        var decoder = new Naming_1.DependencyName(lowerNamingFunction, immutable_1.List([typeName]), function (names) { return Support_1.defined(names.first()); });
        this._namedTypeDependents = this._namedTypeDependents.set(typeName, { encoder: encoder, decoder: decoder });
        return [encoder, decoder];
    };
    Object.defineProperty(ElmRenderer.prototype, "propertyNamer", {
        get: function () {
            return lowerNamingFunction;
        },
        enumerable: true,
        configurable: true
    });
    ElmRenderer.prototype.forbiddenForProperties = function (c, classNamed) {
        return { names: [], namespaces: [this.globalNamespace] };
    };
    Object.defineProperty(ElmRenderer.prototype, "caseNamer", {
        get: function () {
            return upperNamingFunction;
        },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(ElmRenderer.prototype, "casesInGlobalNamespace", {
        get: function () {
            return true;
        },
        enumerable: true,
        configurable: true
    });
    ElmRenderer.prototype.namedTypeToNameForTopLevel = function (type) {
        if (type.isNamedType()) {
            return type;
        }
        return null;
    };
    Object.defineProperty(ElmRenderer.prototype, "arrayType", {
        get: function () {
            return this._useList ? "List" : "Array";
        },
        enumerable: true,
        configurable: true
    });
    ElmRenderer.prototype.emitSourceStructure = function () {
        var _this = this;
        var exports = [];
        var topLevelDecoders = immutable_1.List();
        this.forEachTopLevel("none", function (_, name) {
            var _a = Support_1.defined(_this._topLevelDependents.get(name)), encoder = _a.encoder, decoder = _a.decoder;
            if (decoder === undefined) {
                decoder = Support_1.defined(_this._namedTypeDependents.get(name)).decoder;
            }
            topLevelDecoders = topLevelDecoders.push(decoder);
            exports.push(name, encoder, decoder);
        });
        this.forEachClass("none", function (t, name) {
            if (!_this.topLevels.contains(t))
                exports.push(name);
        });
        this.forEachUnion("none", function (t, name) {
            if (!_this.topLevels.contains(t))
                exports.push([name, "(..)"]);
        });
        this.emitMultiline("-- To decode the JSON data, add this file to your project, run\n--\n--     elm-package install NoRedInk/elm-decode-pipeline\n--\n-- add these imports\n--\n--     import Json.Decode exposing (decodeString)");
        this.emitLine("--     import ", this._moduleName, " exposing (", Support_1.intercalate(", ", topLevelDecoders).toArray(), ")");
        this.emitMultiline("--\n-- and you're off to the races with\n--");
        this.forEachTopLevel("none", function (_, name) {
            var decoder = Support_1.defined(_this._topLevelDependents.get(name)).decoder;
            if (decoder === undefined) {
                decoder = Support_1.defined(_this._namedTypeDependents.get(name)).decoder;
            }
            _this.emitLine("--     decodeString ", decoder, " myJsonString");
        });
        this.emitNewline();
        this.emitLine("module ", this._moduleName, " exposing");
        this.indent(function () {
            for (var i = 0; i < exports.length; i++) {
                _this.emitLine(i === 0 ? "(" : ",", " ", exports[i]);
            }
            _this.emitLine(")");
        });
        this.emitNewline();
        this.emitMultiline("import Json.Decode as Jdec\nimport Json.Decode.Pipeline as Jpipe\nimport Json.Encode as Jenc\nimport Dict exposing (Dict, map, toList)");
        if (this._useList) {
            this.emitLine("import List exposing (map)");
        }
        else {
            this.emitLine("import Array exposing (Array, map)");
        }
        this.forEachTopLevel("leading-and-interposing", this.emitTopLevelDefinition, function (t) { return !_this.namedTypeToNameForTopLevel(t); });
        this.forEachNamedType("leading-and-interposing", false, this.emitClassDefinition, this.emitEnumDefinition, this.emitUnionDefinition);
        this.emitNewline();
        this.emitLine("-- decoders and encoders");
        this.forEachTopLevel("leading-and-interposing", this.emitTopLevelFunctions);
        this.forEachNamedType("leading-and-interposing", false, this.emitClassFunctions, this.emitEnumFunctions, this.emitUnionFunctions);
        this.emitNewline();
        this.emitLine("--- encoder helpers");
        this.emitNewline();
        this.emitLine("make", this.arrayType, "Encoder : (a -> Jenc.Value) -> ", this.arrayType, " a -> Jenc.Value");
        this.emitLine("make", this.arrayType, "Encoder f arr =");
        this.indent(function () {
            _this.emitLine("Jenc.", Strings_1.decapitalize(_this.arrayType), " (", _this.arrayType, ".map f arr)");
        });
        this.emitNewline();
        this.emitMultiline("makeDictEncoder : (a -> Jenc.Value) -> Dict String a -> Jenc.Value\nmakeDictEncoder f dict =\n    Jenc.object (toList (Dict.map (\\k -> f) dict))\n\nmakeNullableEncoder : (a -> Jenc.Value) -> Maybe a -> Jenc.Value\nmakeNullableEncoder f m =\n    case m of\n    Just x -> f x\n    Nothing -> Jenc.null");
    };
    return ElmRenderer;
}(ConvenienceRenderer_1.ConvenienceRenderer));
