#!/usr/bin/env node
"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __generator = (this && this.__generator) || function (thisArg, body) {
    var _ = { label: 0, sent: function() { if (t[0] & 1) throw t[1]; return t[1]; }, trys: [], ops: [] }, f, y, t, g;
    return g = { next: verb(0), "throw": verb(1), "return": verb(2) }, typeof Symbol === "function" && (g[Symbol.iterator] = function() { return this; }), g;
    function verb(n) { return function (v) { return step([n, v]); }; }
    function step(op) {
        if (f) throw new TypeError("Generator is already executing.");
        while (_) try {
            if (f = 1, y && (t = y[op[0] & 2 ? "return" : op[0] ? "throw" : "next"]) && !(t = t.call(y, op[1])).done) return t;
            if (y = 0, t) op = [0, t.value];
            switch (op[0]) {
                case 0: case 1: t = op; break;
                case 4: _.label++; return { value: op[1], done: false };
                case 5: _.label++; y = op[1]; op = [0]; continue;
                case 7: op = _.ops.pop(); _.trys.pop(); continue;
                default:
                    if (!(t = _.trys, t = t.length > 0 && t[t.length - 1]) && (op[0] === 6 || op[0] === 2)) { _ = 0; continue; }
                    if (op[0] === 3 && (!t || (op[1] > t[0] && op[1] < t[3]))) { _.label = op[1]; break; }
                    if (op[0] === 6 && _.label < t[1]) { _.label = t[1]; t = op; break; }
                    if (t && _.label < t[2]) { _.label = t[2]; _.ops.push(op); break; }
                    if (t[2]) _.ops.pop();
                    _.trys.pop(); continue;
            }
            op = body.call(thisArg, _);
        } catch (e) { op = [6, e]; y = 0; } finally { f = t = 0; }
        if (op[0] & 5) throw op[1]; return { value: op[0] ? op[1] : void 0, done: true };
    }
};
Object.defineProperty(exports, "__esModule", { value: true });
var fs = require("fs");
var path = require("path");
var _ = require("lodash");
var _1 = require(".");
var targetLanguages = require("./Language/All");
var URLGrammar_1 = require("./URLGrammar");
var Annotation_1 = require("./Annotation");
var Support_1 = require("./Support");
var commandLineArgs = require("command-line-args");
var getUsage = require("command-line-usage");
var chalk = require("chalk");
var fetch = require("node-fetch");
var langs = targetLanguages.all.map(function (r) { return _.minBy(r.names, function (s) { return s.length; }); }).join("|");
var langDisplayNames = targetLanguages.all.map(function (r) { return r.displayName; }).join(", ");
function sampleFromFileOrUrl(fileOrUrl) {
    return __awaiter(this, void 0, void 0, function () {
        var response;
        return __generator(this, function (_a) {
            switch (_a.label) {
                case 0:
                    if (!fs.existsSync(fileOrUrl)) return [3 /*break*/, 1];
                    return [2 /*return*/, { source: fs.createReadStream(fileOrUrl) }];
                case 1: return [4 /*yield*/, fetch(fileOrUrl)];
                case 2:
                    response = _a.sent();
                    return [2 /*return*/, { source: response.body }];
            }
        });
    });
}
function sourceFromFileOrUrlArray(name, filesOrUrls) {
    return __awaiter(this, void 0, void 0, function () {
        var samples;
        return __generator(this, function (_a) {
            switch (_a.label) {
                case 0: return [4 /*yield*/, Promise.all(filesOrUrls.map(sampleFromFileOrUrl))];
                case 1:
                    samples = _a.sent();
                    return [2 /*return*/, { name: name, samples: samples }];
            }
        });
    });
}
function samplesFromDirectory(dataDir) {
    return __awaiter(this, void 0, void 0, function () {
        function readFilesOrURLsInDirectory(d) {
            return __awaiter(this, void 0, void 0, function () {
                var _this = this;
                var files;
                return __generator(this, function (_a) {
                    files = fs
                        .readdirSync(d)
                        .map(function (x) { return path.join(d, x); })
                        .filter(function (x) { return fs.lstatSync(x).isFile(); });
                    // Each file is a (Name, JSON | URL)
                    return [2 /*return*/, Promise.all(files.map(function (file) { return __awaiter(_this, void 0, void 0, function () {
                            var name, inferredName, fileOrUrl, _a;
                            return __generator(this, function (_b) {
                                switch (_b.label) {
                                    case 0:
                                        name = path.basename(file);
                                        inferredName = name.substr(0, name.lastIndexOf("."));
                                        fileOrUrl = file;
                                        // If file is a URL string, download it
                                        if (_.endsWith(file, ".url")) {
                                            fileOrUrl = fs.readFileSync(file, "utf8").trim();
                                        }
                                        _a = {
                                            inferredName: inferredName
                                        };
                                        return [4 /*yield*/, sampleFromFileOrUrl(fileOrUrl)];
                                    case 1: return [2 /*return*/, (_a.sample = _b.sent(),
                                            _a)];
                                }
                            });
                        }); }))];
                });
            });
        }
        var contents, directories, topLevelSamples, sources, _i, topLevelSamples_1, topLevel, _a, directories_1, dir, samples;
        return __generator(this, function (_b) {
            switch (_b.label) {
                case 0:
                    contents = fs.readdirSync(dataDir).map(function (x) { return path.join(dataDir, x); });
                    directories = contents.filter(function (x) { return fs.lstatSync(x).isDirectory(); });
                    return [4 /*yield*/, readFilesOrURLsInDirectory(dataDir)];
                case 1:
                    topLevelSamples = _b.sent();
                    sources = [];
                    for (_i = 0, topLevelSamples_1 = topLevelSamples; _i < topLevelSamples_1.length; _i++) {
                        topLevel = topLevelSamples_1[_i];
                        sources.push({
                            name: topLevel.inferredName,
                            samples: [topLevel.sample]
                        });
                    }
                    _a = 0, directories_1 = directories;
                    _b.label = 2;
                case 2:
                    if (!(_a < directories_1.length)) return [3 /*break*/, 5];
                    dir = directories_1[_a];
                    return [4 /*yield*/, readFilesOrURLsInDirectory(dir)];
                case 3:
                    samples = _b.sent();
                    sources.push({
                        name: path.basename(dir),
                        samples: samples.map(function (x) { return x.sample; })
                    });
                    _b.label = 4;
                case 4:
                    _a++;
                    return [3 /*break*/, 2];
                case 5: return [2 /*return*/, sources];
            }
        });
    });
}
function inferLang(options) {
    // Output file extension determines the language if language is undefined
    if (options.out) {
        var extension = path.extname(options.out);
        if (extension === "") {
            throw new Error("Please specify a language (--lang) or an output file extension.");
        }
        return extension.substr(1);
    }
    return "go";
}
function inferTopLevel(options) {
    // Output file name determines the top-level if undefined
    if (options.out) {
        var extension = path.extname(options.out);
        var without = path.basename(options.out).replace(extension, "");
        return without;
    }
    // Source determines the top-level if undefined
    if (options.src && options.src.length === 1) {
        var src = options.src[0];
        var extension = path.extname(src);
        var without = path.basename(src).replace(extension, "");
        return without;
    }
    return "TopLevel";
}
function inferOptions(opts) {
    var srcLang = opts.srcLang;
    if (opts.graphqlSchema !== undefined) {
        Support_1.assert(srcLang === undefined || srcLang === "graphql", "If a GraphQL schema is specified, the source language must be GraphQL");
        srcLang = "graphql";
    }
    else {
        Support_1.assert(srcLang !== "graphql", "Please specify a GraphQL schema with --graphql-schema");
        srcLang = srcLang || "json";
    }
    return {
        src: opts.src || [],
        srcLang: srcLang,
        lang: opts.lang || inferLang(opts),
        topLevel: opts.topLevel || inferTopLevel(opts),
        noMaps: !!opts.noMaps,
        noEnums: !!opts.noEnums,
        noCombineClasses: !!opts.noCombineClasses,
        noRender: !!opts.noRender,
        rendererOptions: opts.rendererOptions || {},
        help: opts.help || false,
        quiet: opts.quiet || false,
        out: opts.out,
        graphqlSchema: opts.graphqlSchema
    };
}
var optionDefinitions = [
    {
        name: "out",
        alias: "o",
        type: String,
        typeLabel: "FILE",
        description: "The output file. Determines --lang and --top-level."
    },
    {
        name: "top-level",
        alias: "t",
        type: String,
        typeLabel: "NAME",
        description: "The name for the top level type."
    },
    {
        name: "lang",
        alias: "l",
        type: String,
        typeLabel: langs,
        description: "The target language."
    },
    {
        name: "src-lang",
        alias: "s",
        type: String,
        defaultValue: undefined,
        typeLabel: "json|schema|graphql",
        description: "The source language (default is json)."
    },
    {
        name: "src",
        type: String,
        multiple: true,
        defaultOption: true,
        typeLabel: "FILE|URL|DIRECTORY",
        description: "The file, url, or data directory to type."
    },
    {
        name: "src-urls",
        type: String,
        typeLabel: "FILE",
        description: "Tracery grammar describing URLs to crawl."
    },
    {
        name: "no-combine-classes",
        type: Boolean,
        description: "Don't combine similar classes."
    },
    {
        name: "graphql-schema",
        type: String,
        typeLabel: "FILE",
        description: "GraphQL introspection file."
    },
    {
        name: "no-maps",
        type: Boolean,
        description: "Don't infer maps, always use classes."
    },
    {
        name: "no-enums",
        type: Boolean,
        description: "Don't infer enums, always use strings."
    },
    {
        name: "no-render",
        type: Boolean,
        description: "Don't render output."
    },
    {
        name: "quiet",
        type: Boolean,
        description: "Don't show issues in the generated code."
    },
    {
        name: "help",
        alias: "h",
        type: Boolean,
        description: "Get some help."
    }
];
var sectionsBeforeRenderers = [
    {
        header: "Synopsis",
        content: "$ quicktype [[bold]{--lang} " + langs + "] FILE|URL ..."
    },
    {
        header: "Description",
        content: "Given JSON sample data, quicktype outputs code for working with that data in " + langDisplayNames + "."
    },
    {
        header: "Options",
        optionList: optionDefinitions,
        hide: ["no-render"]
    }
];
var sectionsAfterRenderers = [
    {
        header: "Examples",
        content: [
            chalk.dim("Generate C# to parse a Bitcoin API"),
            "$ quicktype -o LatestBlock.cs https://blockchain.info/latestblock",
            "",
            chalk.dim("Generate Go code from a directory of samples containing:"),
            chalk.dim("  - Foo.json\n  + Bar\n    - bar-sample-1.json\n    - bar-sample-2.json\n  - Baz.url"),
            "$ quicktype -l go samples",
            "",
            chalk.dim("Generate JSON Schema, then TypeScript"),
            "$ quicktype -o schema.json https://blockchain.info/latestblock",
            "$ quicktype -o bitcoin.ts --src-lang schema schema.json"
        ]
    },
    {
        content: "Learn more at [bold]{quicktype.io}"
    }
];
function getOptionDefinitions(opts) {
    return _1.getTargetLanguage(opts.lang).optionDefinitions;
}
function parseArgv(argv) {
    // We can only fully parse the options once we know which renderer is selected,
    // because there are renderer-specific options.  But we only know which renderer
    // is selected after we've parsed the options.  Hence, we parse the options
    // twice.  This is the first parse to get the renderer:
    var incompleteOptions = parseOptions(optionDefinitions, argv, true);
    var rendererOptionDefinitions = getOptionDefinitions(incompleteOptions);
    // Use the global options as well as the renderer options from now on:
    var allOptionDefinitions = _.concat(optionDefinitions, rendererOptionDefinitions);
    try {
        // This is the parse that counts:
        return parseOptions(allOptionDefinitions, argv, false);
    }
    catch (error) {
        if (error.name === "UNKNOWN_OPTION") {
            usage();
            throw new Error("Unknown option");
        }
        throw error;
    }
}
// Parse the options in argv and split them into global options and renderer options,
// according to each option definition's `renderer` field.  If `partial` is false this
// will throw if it encounters an unknown option.
function parseOptions(definitions, argv, partial) {
    var opts = commandLineArgs(definitions, { argv: argv, partial: partial });
    var options = { rendererOptions: {} };
    definitions.forEach(function (o) {
        if (!(o.name in opts))
            return;
        var v = opts[o.name];
        if (o.renderer)
            options.rendererOptions[o.name] = v;
        else {
            var k = _.lowerFirst(o.name
                .split("-")
                .map(_.upperFirst)
                .join(""));
            options[k] = v;
        }
    });
    return inferOptions(options);
}
function usage() {
    var rendererSections = [];
    _.forEach(targetLanguages.all, function (language) {
        var definitions = language.optionDefinitions;
        if (definitions.length === 0)
            return;
        rendererSections.push({
            header: "Options for " + language.displayName,
            optionList: definitions
        });
    });
    var sections = _.concat(sectionsBeforeRenderers, rendererSections, sectionsAfterRenderers);
    console.log(getUsage(sections));
}
function splitAndWriteJava(dir, str) {
    var lines = str.split("\n");
    var filename = null;
    var currentFileContents = "";
    var writeFile = function () {
        if (filename != null) {
            fs.writeFileSync(path.join(dir, filename), currentFileContents);
        }
        filename = null;
        currentFileContents = "";
    };
    var i = 0;
    while (i < lines.length) {
        var line = lines[i];
        i += 1;
        var results = line.match("^// (.+\\.java)$");
        if (results == null) {
            currentFileContents += line + "\n";
        }
        else {
            writeFile();
            filename = results[1];
            while (lines[i] === "")
                i++;
        }
    }
    writeFile();
}
function getSources(options) {
    return __awaiter(this, void 0, void 0, function () {
        var json, jsonMap_1, topLevels, exists, directories_2, sources, _i, directories_3, dataDir, _a, _b, filesOrUrls, _c, _d;
        return __generator(this, function (_e) {
            switch (_e.label) {
                case 0:
                    if (!options.srcUrls) return [3 /*break*/, 1];
                    json = JSON.parse(fs.readFileSync(options.srcUrls, "utf8"));
                    jsonMap_1 = URLGrammar_1.urlsFromURLGrammar(json);
                    topLevels = Object.getOwnPropertyNames(jsonMap_1);
                    return [2 /*return*/, Promise.all(topLevels.map(function (name) { return sourceFromFileOrUrlArray(name, jsonMap_1[name]); }))];
                case 1:
                    if (!(options.src.length === 0)) return [3 /*break*/, 2];
                    return [2 /*return*/, [
                            {
                                name: options.topLevel,
                                samples: [{ source: process.stdin }]
                            }
                        ]];
                case 2:
                    exists = options.src.filter(fs.existsSync);
                    directories_2 = exists.filter(function (x) { return fs.lstatSync(x).isDirectory(); });
                    sources = [];
                    _i = 0, directories_3 = directories_2;
                    _e.label = 3;
                case 3:
                    if (!(_i < directories_3.length)) return [3 /*break*/, 6];
                    dataDir = directories_3[_i];
                    _b = (_a = sources).concat;
                    return [4 /*yield*/, samplesFromDirectory(dataDir)];
                case 4:
                    sources = _b.apply(_a, [_e.sent()]);
                    _e.label = 5;
                case 5:
                    _i++;
                    return [3 /*break*/, 3];
                case 6:
                    filesOrUrls = options.src.filter(function (x) { return !_.includes(directories_2, x); });
                    if (!!_.isEmpty(filesOrUrls)) return [3 /*break*/, 8];
                    _d = (_c = sources).push;
                    return [4 /*yield*/, sourceFromFileOrUrlArray(options.topLevel, filesOrUrls)];
                case 7:
                    _d.apply(_c, [_e.sent()]);
                    _e.label = 8;
                case 8: return [2 /*return*/, sources];
            }
        });
    });
}
function main(args) {
    return __awaiter(this, void 0, void 0, function () {
        var options, sources, _a, queryFile, schemaFile, schema, query, jsonSources, run, _b, lines_1, annotations, output;
        return __generator(this, function (_c) {
            switch (_c.label) {
                case 0:
                    if (!(_.isArray(args) && args.length === 0)) return [3 /*break*/, 1];
                    usage();
                    return [3 /*break*/, 10];
                case 1:
                    options = _.isArray(args) ? parseArgv(args) : inferOptions(args);
                    if (options.help) {
                        usage();
                        return [2 /*return*/];
                    }
                    sources = void 0;
                    _a = options.srcLang;
                    switch (_a) {
                        case "graphql": return [3 /*break*/, 2];
                        case "json": return [3 /*break*/, 3];
                        case "schema": return [3 /*break*/, 5];
                    }
                    return [3 /*break*/, 7];
                case 2:
                    if (options.src.length !== 1) {
                        return [2 /*return*/, Support_1.panic("Please specify one GraphQL query as input.")];
                    }
                    queryFile = options.src[0];
                    schemaFile = Support_1.defined(options.graphqlSchema);
                    schema = JSON.parse(fs.readFileSync(schemaFile, "utf8"));
                    query = fs.readFileSync(queryFile, "utf8");
                    sources = { topLevelName: options.topLevel, schema: schema, query: query };
                    return [3 /*break*/, 8];
                case 3: return [4 /*yield*/, getSources(options)];
                case 4:
                    sources = _c.sent();
                    return [3 /*break*/, 8];
                case 5: return [4 /*yield*/, getSources(options)];
                case 6:
                    jsonSources = _c.sent();
                    sources = jsonSources.map(function (jsonSource) {
                        Support_1.assert(jsonSource.samples.length === 1, "Please specify one schema file for " + jsonSource.name);
                        return {
                            name: jsonSource.name,
                            schema: jsonSource.samples[0].source
                        };
                    });
                    return [3 /*break*/, 8];
                case 7:
                    Support_1.panic("Unsupported source language (" + options.srcLang + ")");
                    return [3 /*break*/, 8];
                case 8:
                    run = new _1.Run({
                        lang: options.lang,
                        sources: sources,
                        inferMaps: !options.noMaps,
                        inferEnums: !options.noEnums,
                        combineClasses: !options.noCombineClasses,
                        noRender: options.noRender,
                        rendererOptions: options.rendererOptions
                    }, false);
                    return [4 /*yield*/, run.run()];
                case 9:
                    _b = _c.sent(), lines_1 = _b.lines, annotations = _b.annotations;
                    output = lines_1.join("\n");
                    if (options.out) {
                        if (options.lang === "java") {
                            splitAndWriteJava(path.dirname(options.out), output);
                        }
                        else {
                            fs.writeFileSync(options.out, output);
                        }
                    }
                    else {
                        process.stdout.write(output);
                    }
                    if (options.quiet) {
                        return [2 /*return*/];
                    }
                    annotations.forEach(function (sa) {
                        var annotation = sa.annotation;
                        if (!(annotation instanceof Annotation_1.IssueAnnotationData))
                            return;
                        var lineNumber = sa.span.start.line;
                        var humanLineNumber = lineNumber + 1;
                        console.error("\nIssue in line " + humanLineNumber + ": " + annotation.message);
                        console.error(humanLineNumber + ": " + lines_1[lineNumber]);
                    });
                    _c.label = 10;
                case 10: return [2 /*return*/];
            }
        });
    });
}
exports.main = main;
if (require.main === module) {
    main(process.argv.slice(2)).catch(function (e) {
        if (e instanceof Error) {
            console.error("Error: " + e.message + ".");
        }
        else {
            console.error(e);
        }
        process.exit(1);
    });
}
